@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2001, 2003-2015, 2017-2019, 2021,
+ * Copyright (C) 1986, 1988, 1989, 1991-2001, 2003-2015, 2017-2019, 2021, 2022,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -25,7 +25,6 @@
  */
 
 #include "awk.h"
-#include "floatmagic.h"	/* definition of isnan */
 
 static NODE *r_make_number(double x);
 static AWKNUM get_ieee_magic_val(char *val);
@@ -61,6 +60,15 @@ r_force_number(NODE *n)
 	char save;
 	char *ptr;
 
+	if (n->type == Node_elem_new) {
+		n->type = Node_val;
+		n->flags &= ~STRING;
+		n->stptr[0] = '0';	// STRCUR is still set
+		n->stlen = 1;
+
+		return n;
+	}
+
 	if ((n->flags & NUMCUR) != 0)
 		return n;
 
@@ -370,7 +378,7 @@ int
 cmp_awknums(const NODE *t1, const NODE *t2)
 {
 	/*
-	 * This routine is also used to sort numeric array indices or values.
+	 * This routine is used to sort numeric array indices or values.
 	 * For the purposes of sorting, NaN is considered greater than
 	 * any other value, and all NaN values are considered equivalent and equal.
 	 * This isn't in compliance with IEEE standard, but compliance w.r.t. NaN
@@ -390,7 +398,6 @@ cmp_awknums(const NODE *t1, const NODE *t2)
 	return 1;
 }
 
-
 /* make_str_node --- make a string node */
 
 NODE *
@@ -1083,3 +1090,24 @@ more_blocks(int id)
 }
 
 #endif
+
+/* make_bool_node --- make a boolean-valued node */
+
+extern NODE *
+make_bool_node(bool value)
+{
+	NODE *val;
+	const char *sval;
+	AWKNUM nval;
+
+	sval = (value ? "1" : "0");
+	nval = (value ? 1.0 : 0.0);
+
+	val = make_number(nval);
+	val->stptr = estrdup(sval, strlen(sval));
+	val->stlen = strlen(sval);
+	val->flags |= NUMCUR|STRCUR|BOOLVAL;
+	val->stfmt = STFMT_UNUSED;
+
+	return val;
+}
