@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2015, 2017, 2018,
+ * Copyright (C) 1986, 1988, 1989, 1991-2015, 2017-2020, 2022,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -31,14 +31,12 @@ extern INSTRUCTION *rule_list;
 
 #define HASHSIZE	1021
 
-static int func_count;	/* total number of functions */
-static int var_count;	/* total number of global variables and functions */
-
 static NODE *symbol_list;
 static void (*install_func)(NODE *) = NULL;
 static NODE *make_symbol(const char *name, NODETYPE type);
 static NODE *install(const char *name, NODE *parm, NODETYPE type);
 static void free_bcpool(INSTRUCTION_POOL *pl);
+static NODE *get_name_from_awk_ns(const char *name);
 
 static AWK_CONTEXT *curr_ctxt = NULL;
 static int ctxt_level;
@@ -49,10 +47,10 @@ NODE *symbol_table, *func_table;
 /* Use a flag to avoid a strcmp() call inside install() */
 static bool installing_specials = false;
 
-/* init_symbol_table --- make sure the symbol tables are initialized */
+/* init_the_tables --- deal with the tables for in memory use */
 
-void
-init_symbol_table()
+static void
+init_the_tables(void)
 {
 	getnode(global_table);
 	memset(global_table, '\0', sizeof(NODE));
@@ -64,11 +62,56 @@ init_symbol_table()
 
 	installing_specials = true;
 	func_table = install_symbol(estrdup("FUNCTAB", 7), Node_var_array);
-
 	symbol_table = install_symbol(estrdup("SYMTAB", 6), Node_var_array);
 	installing_specials = false;
 }
 
+/* init_symbol_table --- make sure the symbol tables are initialized */
+
+void
+init_symbol_table()
+{
+	if (! using_persistent_malloc) {
+		// normal case, initialize regularly, return
+		init_the_tables();
+		return;
+	}
+
+	// using persistent memory, get the root pointer
+	// which holds this struct:
+	struct root_pointers {
+		NODE *global_table;
+		NODE *func_table;
+		NODE *symbol_table;
+	} *root_pointers = NULL;
+
+	root_pointers = (struct root_pointers *) pma_get_root();
+
+	if (root_pointers == NULL) {
+		// very first time!
+
+		// set up the tables
+		init_the_tables();
+
+		// save the pointers for the next time.
+		emalloc(root_pointers, struct root_pointers *, sizeof(struct root_pointers), "init_symbol_table");
+		root_pointers->global_table = global_table;
+		root_pointers->func_table = func_table;
+		root_pointers->symbol_table = symbol_table;
+		pma_set_root(root_pointers);
+	} else {
+		// this is the next time, get the saved pointers and put them back in place
+		global_table = root_pointers->global_table;
+		func_table = root_pointers->func_table;
+		symbol_table = root_pointers->symbol_table;
+
+		// still need to set this one up as usual
+		getnode(param_table);
+		memset(param_table, '\0', sizeof(NODE));
+		null_array(param_table);
+	}
+}
+
 /*
  * install_symbol:
  * Install a global name in the symbol table, even if it is already there.
@@ -102,11 +145,11 @@ lookup(const char *name)
 	tables[3] = symbol_table;	/* then globals */
 	tables[4] = NULL;
 
-	tmp = make_string(name, strlen(name));
+	tmp = get_name_from_awk_ns(name);
 
 	n = NULL;
 	for (i = 0; tables[i] != NULL; i++) {
-		if (tables[i]->table_size == 0)
+		if (assoc_empty(tables[i]))
 			continue;
 
 		if ((do_posix || do_traditional) && tables[i] == global_table)
@@ -289,6 +332,7 @@ make_symbol(const char *name, NODETYPE type)
 		r->var_value = dupnode(Nnull_string);
 	r->vname = (char *) name;
 	r->type = type;
+	r->valref = 1;
 
 	return r;
 }
@@ -299,12 +343,12 @@ static NODE *
 install(const char *name, NODE *parm, NODETYPE type)
 {
 	NODE *r;
-	NODE **aptr;
 	NODE *table;
 	NODE *n_name;
 	NODE *prev;
 
-	n_name = make_string(name, strlen(name));
+	n_name = get_name_from_awk_ns(name);
+
 	table = symbol_table;
 
 	if (type == Node_param_list) {
@@ -322,10 +366,6 @@ install(const char *name, NODE *parm, NODETYPE type)
 	else {
 		/* global symbol */
 		r = make_symbol(name, type);
-		if (type == Node_func)
-			func_count++;
-		if (type != Node_ext_func && type != Node_builtin_func && table != global_table)
-			var_count++;	/* total, includes Node_func */
 	}
 
 	if (type == Node_param_list) {
@@ -334,14 +374,12 @@ install(const char *name, NODE *parm, NODETYPE type)
 			goto simple;
 		r->dup_ent = prev->dup_ent;
 		prev->dup_ent = r;
+		unref(n_name);
 	} else {
 simple:
 		/* the simple case */
-		aptr = assoc_lookup(table, n_name);
-		unref(*aptr);
-		*aptr = r;
+		assoc_set(table, n_name, r);
 	}
-	unref(n_name);
 
 	if (install_func)
 		(*install_func)(r);
@@ -362,7 +400,18 @@ comp_symbol(const void *v1, const void *v2)
 	n1 = *npp1;
 	n2 = *npp2;
 
-	return strcmp(n1->vname, n2->vname);
+	// names in awk namespace come out first
+	bool n1_is_in_ns = (strchr(n1->vname, ':') != NULL);
+	bool n2_is_in_ns = (strchr(n2->vname, ':') != NULL);
+
+	if (n1_is_in_ns && n2_is_in_ns)
+		return strcmp(n1->vname, n2->vname);
+	else if (n1_is_in_ns && ! n2_is_in_ns)
+		return 1;
+	else if (! n1_is_in_ns && n2_is_in_ns)
+		return -1;
+	else
+		return strcmp(n1->vname, n2->vname);
 }
 
 
@@ -393,7 +442,7 @@ get_symbols(SYMBOL_TYPE what, bool sort)
 		max = the_table->table_size * 2;
 
 		list = assoc_list(the_table, "@unsorted", ASORTI);
-		emalloc(table, NODE **, (func_count + 1) * sizeof(NODE *), "get_symbols");
+		emalloc(table, NODE **, (the_table->table_size + 1) * sizeof(NODE *), "get_symbols");
 
 		for (i = count = 0; i < max; i += 2) {
 			r = list[i+1];
@@ -410,7 +459,7 @@ get_symbols(SYMBOL_TYPE what, bool sort)
 
 		list = assoc_list(the_table, "@unsorted", ASORTI);
 		/* add three: one for FUNCTAB, one for SYMTAB, and one for a final NULL */
-		emalloc(table, NODE **, (var_count + 1 + 1 + 1) * sizeof(NODE *), "get_symbols");
+		emalloc(table, NODE **, (the_table->table_size + 1 + 1 + 1) * sizeof(NODE *), "get_symbols");
 
 		for (i = count = 0; i < max; i += 2) {
 			r = list[i+1];
@@ -468,6 +517,8 @@ print_vars(NODE **table, int (*print_func)(FILE *, const char *, ...), FILE *fp)
 			print_func(fp, "untyped variable\n");
 		else if (r->type == Node_var)
 			valinfo(r->var_value, print_func, fp);
+		else
+			cant_happen("unexpected node type: %s", nodetype2str(r->type));
 	}
 }
 
@@ -593,7 +644,10 @@ load_symbols()
 			    || r->type == Node_var
 			    || r->type == Node_var_array
 			    || r->type == Node_var_new) {
-				tmp = make_string(r->vname, strlen(r->vname));
+				if (strncmp(r->vname, "awk::", 5) == 0)
+					tmp = make_string(r->vname + 5, strlen(r->vname) - 5);
+				else
+					tmp = make_string(r->vname, strlen(r->vname));
 				aptr = assoc_lookup(sym_array, tmp);
 				unref(tmp);
 				unref(*aptr);
@@ -617,7 +671,7 @@ load_symbols()
 					*aptr = dupnode(untyped);
 					break;
 				default:
-					cant_happen();
+					cant_happen("unexpected node type %s", nodetype2str(r->type));
 					break;
 				}
 			}
@@ -630,6 +684,7 @@ load_symbols()
 	unref(scalar);
 	unref(untyped);
 	unref(array);
+	unref(built_in);
 }
 
 /* check_param_names --- make sure no parameter is the name of a function */
@@ -644,7 +699,7 @@ check_param_names(void)
 	bool result = true;
 	NODE n;
 
-	if (func_table->table_size == 0)
+	if (assoc_empty(func_table))
 		return result;
 
 	max = func_table->table_size * 2;
@@ -684,7 +739,7 @@ check_param_names(void)
 
 			if (in_array(func_table, & n)) {
 				error(
-			_("function `%s': can't use function `%s' as a parameter name"),
+			_("function `%s': cannot use function `%s' as a parameter name"),
 					list[i]->stptr,
 					f->fparms[j].param);
 				result = false;
@@ -814,7 +869,7 @@ pop_context()
 
 	assert(curr_ctxt != NULL);
 	if (curr_ctxt->prev == NULL)
-		fatal(_("can not pop main context"));
+		fatal(_("cannot pop main context"));
 	ctxt = curr_ctxt->prev;
 	/* restore source and sourceline */
 	sourceline = ctxt->sourceline;
@@ -907,7 +962,7 @@ free_bc_internal(INSTRUCTION *cp)
 			unref(m);
 		break;
 	case Op_illegal:
-		cant_happen();
+		cant_happen("unexpected opcode %s", opcode2str(cp->opcode));
 	default:
 		break;
 	}
@@ -946,3 +1001,44 @@ free_bcpool(INSTRUCTION_POOL *pl)
 	for (i = 0; i < MAX_INSTRUCTION_ALLOC; i++)
 		free_bc_mempool(& pl->pool[i], i + 1);
 }
+
+/* is_all_upper --- return true if name is all uppercase letters */
+
+/*
+ * DON'T use isupper(), it's locale aware!
+ */
+
+bool
+is_all_upper(const char *name)
+{
+	for (; *name != '\0'; name++) {
+		switch (*name) {
+		case 'A': case 'B': case 'C': case 'D': case 'E':
+		case 'F': case 'G': case 'H': case 'I': case 'J':
+		case 'K': case 'L': case 'M': case 'N': case 'O':
+		case 'P': case 'Q': case 'R': case 'S': case 'T':
+		case 'U': case 'V': case 'W': case 'X': case 'Y':
+		case 'Z':
+			break;
+		default:
+			return false;
+		}
+	}
+
+	return true;
+}
+
+/* get_name_from_awk_ns --- get the name after awk:: or the full name */
+
+static NODE *
+get_name_from_awk_ns(const char *name)
+{
+	NODE *tmp;
+
+	if (strncmp(name, "awk::", 5) == 0)
+		tmp = make_string(name + 5, strlen(name) - 5);
+	else
+		tmp = make_string(name, strlen(name));
+
+	return tmp;
+}
